% Este modulo fornece uma interface entre o modulo Resolvedor e o modulo
% Nucleo.
functor

export
   % Funcao que recebe os dados de entrada de um problema, no formato utilizado
   % pelo modulo Resolvedor, e devolve uma lista com os dados de entrada no
   % formato utilizado internamente pelo modulo Nucleo.
   AdaptaAoScript

define

   % Adapta todos os dados de entrada para formatos mais adequados ao
   % script.
   fun {AdaptaAoScript NumDias TurmasEntrada GruposTurmas
        PreferenciasEntrada RestrProfEntrada Foco UnidadesPorDia
        TamanhoUnidade}
      % Implementa os seguintes procedimentos auxiliares:
      % AdaptaAulaAoScript, AdaptaPrefAoScript, AdaptaRestrProfAoScript,
      % AvaliaTurma, ComparaAvaliacoes.

      % Recebe:
      % - Uma aula, no formato dado pela entrada.
      % - O identificador da turma correspondente a aula.
      % Devolve uma aula, no formato adequado ao modulo Nucleo (a lista de
      % exclusao mutua nao e preenchida).
      fun {AdaptaAulaAoScript AulaEntrada IdTurma}
         % Horarios em que a aula pode ser dada
         Horarios =
         {List.foldL AulaEntrada.hor
          fun {$ Atual D#H#M}
             HorNaSemana = ((D-1)*UnidadesPorDia) +
             ((60*H+M) div TamanhoUnidade)
          in
             {List.append Atual [HorNaSemana] $}
          end
          nil $
         }
      in
         aula(turma: IdTurma dur:(AulaEntrada.dur div TamanhoUnidade)
              hor:Horarios exclMut:_)
      end

      % Recebe uma preferencia, no formato dado pela entrada, e devolve
      % a preferencia no formato adequado ao modulo Nucleo.
      % Os campos dificuldade e numDias sao criados apenas se o foco estiver
      % na solucao otima. De qualquer forma, esses campos nao sao preenchidos
      % por esta funcao.
      fun {AdaptaPrefAoScript PrefEntrada}
         % Intervalos indesejados, nesta preferencia
         Indesejados =
         {List.foldL PrefEntrada.indesejados
          fun {$ Atual D#HI#HF}
             DiaParaUnidades = ((D-1)*UnidadesPorDia)
             HorInicialNaSemana = DiaParaUnidades +
             ((60*HI) div TamanhoUnidade)
             HorFinalNaSemana = DiaParaUnidades + ((60*HF) div TamanhoUnidade)
          in
             {List.append Atual [HorInicialNaSemana#HorFinalNaSemana] $}
          end
          nil $
         }
      in
         if Foco == otima then
            pref(prof:PrefEntrada.prof indesejados:Indesejados
                 peso:PrefEntrada.peso dificuldade:_ numDias:_)
         else
            pref(prof:PrefEntrada.prof indesejados:Indesejados
                 peso:PrefEntrada.peso)
         end
      end

      % Recebe as restricoes de horarios de um professor e devolve as
      % restricoes nom formato adequado ao modulo Nucleo.
      fun {AdaptaRestrProfAoScript RestrEntrada}
         % Intervalos em que este professor nao pode dar aulas
         HorariosImpossiveis =
         {List.foldL RestrEntrada.impossiveis
          fun {$ Atual D#HI#HF}
             DiaParaUnidades = ((D-1)*UnidadesPorDia)
             HorInicialNaSemana = DiaParaUnidades +
             ((60*HI) div TamanhoUnidade)
             HorFinalNaSemana = DiaParaUnidades + ((60*HF) div TamanhoUnidade)
          in
             {List.append Atual [HorInicialNaSemana#HorFinalNaSemana] $}
          end
          nil $
         }
      in
         restrProf(prof:RestrEntrada.prof impossiveis:HorariosImpossiveis)
      end

      % Recebe:
      % - IdTurma#Turma, onde IdTurma e o identificador de uma turma e Turma
      % e a turma correspondente.
      % - TodosOsGrupos: os grupos de exclusao mutua de turmas, incluindo
      % os conjuntos de turmas de cada professor.
      % Devolve uma avaliacao para a turma, na forma Dic#NumExclMut, onde:
      % - Dic e um dicionario em que cada chave C e o tamanho de um grupo de
      % exclusao mutua ao qual esta turma pertence e cada item (correspondente
      % a uma chave C) e a quantidade de grupos de tamanho C aos quais
      % a aula pertence.
      % - NumExclMut e o numero de turmas que estao na lista de exclusao mutua
      % de Turma.
      fun {AvaliaTurma IdTurma#Turma TodosOsGrupos}
         DicExclMut = {Dictionary.new}
      in
         % Percorre os grupos, preenchendo DicExclMut
         {List.forAll TodosOsGrupos
          proc {$ Grupo}
             if {List.member IdTurma Grupo} then
                TamGrupo = {List.length Grupo}
                QuantAtual = {Dictionary.condGet DicExclMut TamGrupo 0}
             in
                {Dictionary.put DicExclMut TamGrupo QuantAtual+1}
             end
          end
         }
         DicExclMut#{List.length Turma.exclMut}
      end

      % Recebe:
      % As avaliacoes de duas turmas.
      % Devolve true se a primeira avaliacao for pior que a segunda, ou seja,
      % se a turma representada pela avaliacao Dic1#Num1 for mais "critica"
      % que a turma representada por Dic2#Num2. Caso contrario, devolve false.
      fun {ComparaAvaliacoes Dic1#Num1 Dic2#Num2}
         % Compara duas listas que representam as entradas de um dicionario
         % de avaliacao.
         % Devolve:
         % - ~1, se a primeira lista representar uma aula mais "critica" que
         % a segunda.
         % - 1, se a segunda lista representar uma aula mais "critica" que a
         % segunda.
         % - 0, se as listas forem iguais.
         fun {ComparaListas L1 L2}
            % Se uma das listas e nula e a outra nao, a lista que nao e nula
            % representa uma aula mais "critica".
            if L1 == nil then
               if L2 == nil then 0 else 1 end
            else
               if L2 == nil then ~1
               else H1|T1 = L1 H2|T2 = L2 in
                  if H1.1 == H2.1 then
                     if H1.2 == H2.2 then {ComparaListas T1 T2}
                     else
                        if H1.2 > H2.2 then ~1 else 1 end
                     end
                  else if H1.1 > H2.1 then ~1 else 1 end
                  end
               end
            end
         end
         MaiorTam = fun {$ T1#_ T2#_} T1 > T2 end
         % Cria listas com as entradas dos dicionarios das avaliacoes.
         ListaExclMut1 = {List.sort {Dictionary.entries Dic1} MaiorTam}
         ListaExclMut2 = {List.sort {Dictionary.entries Dic2} MaiorTam}
         % Compara as listas.
         ResultComp = {ComparaListas ListaExclMut1 ListaExclMut2}
      in
         % Se, comparando os dicionarios, as turmas forem igualmente criticas,
         % considera-se mais critica a turma que possui maior lista de exclusao
         % mutua.
         if ResultComp == 0 then
            Num1 >= Num2
         else ResultComp == ~1 end
      end
      % FIM DO PROCEDIMENTO ComparaAvaliacoes %

      % DADOS DE SAIDA %
      Aulas NumAulas Turmas = {Dictionary.new}
      Grupos Professores = {Dictionary.new}
      Preferencias RestrProfessores
      MatrizExclMut

      % Dicionario, indexado por professores, contendo, para cada professor,
      % a lista com os identificadores de suas turmas.
      TurmasPorProf = {Dictionary.new}
      % Numero de aulas.
      NumAulas
      % Dicionario contendo uma avaliacao para cada turma, indicando o grau
      % de interferencia com as outras turmas do problema.
      DifTurmas = {Dictionary.new}
      % As aulas serao, inicialmente, guardadas numa lista, da qual,
      % posteriormente, sera construida uma tupla.
      ListaDeAulas
      % Lista de preferencias, no formato adequado a modulo Nucleo, mas na
      % ordem dada pela entrada
      PrefsDesord
      % Grupos de exclusao mutua, incluindo os conjuntos das turmas de cada
      % professor.
      TodosOsGrupos
      % Lista de turmas da entrada, em ordem decrescente de avaliacao,
      % no sentido de que as turmas mais "criticas" (que mais interferem
      % em outras turmas) aparecem nas primeiras posicoes da lista.
      TurmasOrd

   in
      % PREENCHE O DICIONARIO TurmasPorProf %
      {ForAll {Dictionary.entries TurmasEntrada}
       proc {$ IdTurmaEntrada#TurmaEntrada}
          % Professor desta turma
          Prof = TurmaEntrada.prof
          % Lista atual de turmas deste professor
          TurmasDoProf = {Dictionary.condGet TurmasPorProf Prof nil}
       in
          {Dictionary.put TurmasPorProf Prof {List.append TurmasDoProf
                                              [IdTurmaEntrada]}}
       end
      }

      % ORDENACAO DAS TURMAS %

      % Cria a lista de grupos de exclusao mutua de turmas, incluindo
      % os conjuntos de turmas de cada professor
      TodosOsGrupos =
      {List.append GruposTurmas {Dictionary.items TurmasPorProf}}

      % Associa a cada turma uma avaliacao, indicando o grau de
      % interferencia com as outras turmas do problema.
      {List.forAll {Dictionary.entries TurmasEntrada}
       proc {$ IdTurma#Turma}
          {Dictionary.put DifTurmas IdTurma
           {AvaliaTurma IdTurma#Turma TodosOsGrupos}}
       end
      }

      % Cria a lista ordenada de turmas
      TurmasOrd =
      {List.sort {Dictionary.entries TurmasEntrada}
       fun {$ Id1#_ Id2#_}
          {ComparaAvaliacoes {Dictionary.get DifTurmas Id1}
           {Dictionary.get DifTurmas Id2}}
       end
      }

      % CRIA UMA LISTA DE AULAS E PREENCHE O DICIONARIO DE TURMAS %
      % (Observe que a ordem das aulas sera influenciada pela ordem das turmas)
      {List.foldL TurmasOrd
       fun {$ ListaAtualDeAulas IdTurma#Turma}
          AulasDaTurmaAtual
          % Cria um registro para a turma, sem ainda definir as aulas
          TurmaAtual = turma(aulas:AulasDaTurmaAtual dist:Turma.dist
                             horVaria:Turma.horVaria)
       in
          % Inclui o registro no dicionario de turmas
          {Dictionary.put Turmas IdTurma TurmaAtual}
          % Define os numeros das aulas da turma atual e acrescenta
          % essas aulas a lista geral de aulas
          {List.foldL Turma.aulas
           fun {$ ListaAtual#AtualDaTurma Aula}
              Result1 = {List.append ListaAtual
                         [{AdaptaAulaAoScript Aula IdTurma}]}
              Result2 = {List.append AtualDaTurma [{List.length Result1}]}
           in
              Result1#Result2
           end
           ListaAtualDeAulas#nil $#AulasDaTurmaAtual
          }
       end
       nil ListaDeAulas
      }
      % CRIA A TUPLA DE AULAS, A PARTIR DA LISTA %
      Aulas = {List.toTuple aulas ListaDeAulas}

      % CRIA A MATRIZ DE EXCLUSAO MUTUA

      % Obtem o numero total de aulas
      NumAulas = {Record.width Aulas}

      % Cria a matriz
      MatrizExclMut = {Array.new 1 NumAulas _}
      % Por enquanto, nao ha exclusao mutua e, portanto, todas as posicoes
      % da matriz contem o valor false.
      {For 1 NumAulas 1
       proc {$ I}
          {Array.put MatrizExclMut I {Array.new 1 NumAulas false}}
       end
      }

      % DEFINE A LISTA DE EXCLUSAO MUTUA DE CADA AULA %
      % Percorre as turmas...
      {List.forAll {Dictionary.entries Turmas}
       proc {$ IdTurma#Turma}
          TurmasExclMut = {Dictionary.get TurmasEntrada IdTurma}.exclMut
          % Aulas de outras turmas que tem exclusao mutua com as aulas
          % desta turma
          AulasExclMut =
          if TurmasExclMut \= nil then
             {List.foldL TurmasExclMut
              fun {$ Atual IdTurmaExcl}
                 {List.append Atual
                  {Dictionary.get Turmas IdTurmaExcl}.aulas $}
              end
              nil $
             }
          else nil end
       in
          % Compoe a lista de exclusao mutua de cada aula desta turma com:
          % - As aulas de AulasExclMut
          % - As outras aulas desta turma
          % Aualiza as posicoes correspondentes da matriz de exclusao mutua
          {List.forAllTail Turma.aulas
           proc {$ Tail}
              Primeira|Resto = Tail
           in
              if Resto == nil then
                 Aulas.Primeira.exclMut = AulasExclMut
              else
                 Aulas.Primeira.exclMut = {List.append AulasExclMut Resto}
              end
              % Atualizacao da matriz
              {List.forAll Aulas.Primeira.exclMut
               proc {$ AEM}
                  {Array.put {Array.get MatrizExclMut Primeira} AEM true}
                  {Array.put {Array.get MatrizExclMut AEM} Primeira true}
               end
              }
           end
          }
       end
      }
      % CRIA A LISTA DE GRUPOS DE AULAS PARA EXCLUSAO MUTUA %
      if GruposTurmas \= nil then
         % Percorre os grupos de turmas, criando os grupos de aulas
         % correspondentes.
         {List.foldL GruposTurmas
          fun {$ ListaAtual GrupoTurmas}
             GrupoAulas =
             {List.foldL GrupoTurmas
              fun {$ Atual IdTurma}
                 {List.append Atual {Dictionary.get Turmas IdTurma}.aulas}
              end
              nil $
             }
          in
             {List.append ListaAtual [GrupoAulas] $}
          end
          nil Grupos
         }
      else Grupos = nil end
      % Atualiza a matriz de exclusao mutua
      {List.forAll Grupos
       proc {$ G}
          {List.forAllTail G
           proc {$ H|T}
              {List.forAll T
               proc {$ AEM}
                  {Array.put {Array.get MatrizExclMut H} AEM true}
                  {Array.put {Array.get MatrizExclMut AEM} H true}
               end
              }
           end
          }
       end
      }
      % PREENCHE O DICIONARIO Professores %
      {ForAll {Dictionary.entries TurmasEntrada}
       proc {$ IdTurmaEntrada#TurmaEntrada}
          % Professor dessa turma
          Prof = TurmaEntrada.prof
          % Lista com os numeros das aulas dessa turma
          AulasDaTurma = {Dictionary.get Turmas IdTurmaEntrada}.aulas
          % Lista atual de aulas desse professor (pode ser nil)
          AulasDoProf = {Dictionary.condGet Professores Prof nil}
       in
          {Dictionary.put Professores Prof
           {List.append AulasDoProf AulasDaTurma}}
       end
      }
      % Atualiza a matriz de exclusao mutua
      {List.forAll {Dictionary.items Professores}
       proc {$ ListaDeAulas}
          {List.forAllTail ListaDeAulas
           proc {$ H|T}
              {List.forAll T
               proc {$ Num}
                  {Array.put {Array.get MatrizExclMut H} Num true}
                  {Array.put {Array.get MatrizExclMut Num} H true}
               end
              }
           end
          }
       end
      }

      % CRIA A LISTA DE PREFERENCIAS %
      if PreferenciasEntrada \= nil then
         {List.foldL PreferenciasEntrada
          fun {$ Atual Preferencia}
             {List.append Atual [{AdaptaPrefAoScript Preferencia}]}
          end
          nil PrefsDesord
         }
      else PrefsDesord = nil end

      % SE O FOCO ESTIVER NA SOLUCAO OTIMA, SAO FEITAS A AVALIACAO
      % E A ORDENACAO DAS PREFERENCIAS.

      if Foco == otima then
         % CALCULA OS CAMPOS "numDias" e "dificuldade" de cada preferencia.
         {List.forAll PrefsDesord
          proc {$ Pref}
             % Numeros das aulas do professor que tem esta preferencia
             AulasDoProf = {Dictionary.get Professores Pref.prof}
             % Quantidade de aulas do professor
             NumAulasDoProf = {List.length AulasDoProf}
             % Lista dos numeros das aulas afetadas por esta preferencia
             AulasAfetadas = {Cell.new nil}
             % A posicao I sera vinculada a true se a preferencia envolver
             % o dia I e este dia afetar alguma aula.
             DiasPresentes = {Tuple.make dias NumDias}
             % Valor inicial para a dificuldade desta preferencia (a ser
             % calculada). O motivo deste valor inicial e o fato de que as
             % aulas de um professor formam um grupo de exclusao mutua com
             % NumAulasDoProf aulas.
             Dificuldade = {Cell.new NumAulasDoProf#NumAulasDoProf}
             % Lista de aulas que pertencem a grupos de exclusao mutua com
             % o tamanho TamExclMax corrente.
             AulasCriticas = {Cell.new nil}
          in
             % Verifica, para cada intervalo da preferencia, se alguma aula
             % e afetada pelo intervalo.
             {List.forAll Pref.indesejados
              proc {$ I#F}
                 % Para cada aula, verifica se ela e afetada por este intervalo
                 {List.forAll AulasDoProf
                  proc {$ NumAula}
                     Aula = Aulas.NumAula
                     Afeta =
                     {List.foldL Aula.hor
                      fun {$ ResAtual Horario}
                         ResAtual orelse
                         (Horario < F andthen Horario+Aula.dur > I)
                      end
                      false $
                     }
                  in
                     if Afeta then
                        % O dia em que esta o intervalo atual
                        D = (I div UnidadesPorDia)+1
                     in
                        % Marca D como um dia que afeta alguma aula
                        if {Value.isFree DiasPresentes.D} then
                           DiasPresentes.D = true
                        end
                        % Acrescenta esta aula a lista de aulas afetadas
                        {Cell.assign AulasAfetadas
                         NumAula|{Cell.access AulasAfetadas}}
                     end
                  end
                 }
              end
             }
             % CALCULO DO NUMERO DE DIAS %
             Pref.numDias =
             {Record.foldL DiasPresentes
              fun {$ Soma Valor}
                 if {Value.isDet Valor} then Soma+1 else Soma end
              end
              0 $
             }
             % CALCULO DA DIFICULDADE %
             % Percorre os grupos de exclusao mutua, verificando as aulas
             % que pertencem a esta preferencia.
             {List.forAll Grupos
              proc {$ Grupo}
                 % Tamanho do grupo atual.
                 TamGrupo = {List.length Grupo}
                 % Tamanho corrente do maior grupo de exclusao mutua que contem
                 % alguma aula afetada por esta preferencia.
                 TA = {Cell.access Dificuldade}.1
              in
                 % Se o maior grupo tiver tamanho igual a NumAulasDoProf, entao
                 % nao e necessario calcular o numero de aulas que pertencem
                 % a grupos desse tamanho, pois sabemos que e NumAulasDoProf.
                 if TamGrupo > NumAulasDoProf
                    % Se houver mais de um grupo com tamanho TamExclMax,
                    % devem ser contadas todas as aulas que aparecem nesses
                    % grupos
                    andthen TamGrupo >= TA then

                    % Percorre as aulas do grupo atual...
                    {List.forAll Grupo
                     proc {$ NumAula}
                        % A aula so e considerada se for afetada
                        if {List.member NumAula {Cell.access AulasAfetadas}}
                        then
                           TaAt#NuAt = {Cell.access Dificuldade}
                        in
                           % Se existe mais de um grupo com o tamanho
                           % TamExclMax corrente, devemos considerar tambem as
                           % aulas desse grupo.
                           if TamGrupo == TaAt then
                              ListaCorrente = {Cell.access AulasCriticas}
                           in
                              if {List.member NumAula ListaCorrente} == false
                              then {Cell.assign AulasCriticas
                                    NumAula|ListaCorrente}
                              end
                              {Cell.assign Dificuldade TaAt#(NuAt+1)}
                           else % O tamanho TamExclMax deve mudar
                              {Cell.assign AulasCriticas [NumAula]}
                              {Cell.assign Dificuldade TamGrupo#1}
                           end
                        end
                     end
                    }
                 end
              end
             } % Fim do calculo da dificuldade %
             Pref.dificuldade = {Cell.access Dificuldade}
          end
         }
         % ORDENA AS PREFERENCIAS %
         Preferencias =
         {List.sort PrefsDesord
          fun {$ Pref1 Pref2}
             if Pref1.dificuldade.1 == Pref2.dificuldade.1 then
                if Pref1.dificuldade.2 == Pref2.dificuldade.2 then
                   if Pref1.numDias == Pref2.numDias then
                      Pref1.peso >= Pref2.peso
                   else
                      Pref1.numDias < Pref2.numDias
                   end
                else
                   Pref1.dificuldade.2 < Pref2.dificuldade.2
                end
             else
                Pref1.dificuldade.1 < Pref2.dificuldade.1
             end
          end
         }
      else
         Preferencias = PrefsDesord
      end
      % CRIA A LISTA DE RESTRICOES DE HORARIOS DOS PROFESSORES %
      if RestrProfEntrada \= nil then
         {List.foldL RestrProfEntrada
          fun {$ Atual Restricao}
             {List.append Atual [{AdaptaRestrProfAoScript Restricao}]}
          end
          nil RestrProfessores
         }
      else RestrProfessores = nil end
      [Aulas Turmas Grupos Professores Preferencias RestrProfessores
       MatrizExclMut]
   end
   %% FIM DO PROCEDIMENTO AdaptaAoScript %%
end